#!/usr/bin/env python

# Copyright 2020 Lucas Heitzmann Gabrielli.
# This file is part of gdstk, distributed under the terms of the
# Boost Software License - Version 1.0.  See the accompanying
# LICENSE file or <http://www.boost.org/LICENSE_1_0.txt>

import pathlib
import numpy
import gdstk
from tutorial_images import draw


def init_image():
    polygon_1 = gdstk.Polygon([(0, 0), (1, 0), 1 + 1.5j, 1j])
    polygon_2 = gdstk.Polygon([0j, (-1, 1), (-1, 0)], 2, 2)
    return gdstk.Cell("init").add(polygon_1, polygon_2)


def bounding_box_image():
    polygon = gdstk.Polygon([(0, 1), (1, 2), (3, -1)])
    bbox = polygon.bounding_box()
    polygon_bb = gdstk.rectangle(*bbox, datatype=1)
    return gdstk.Cell("bounding_box").add(polygon, polygon_bb)


def fillet_image():
    points = [(0, 0), (1.2, 0), (1.2, 0.3), (1, 0.3), (1.5, 1), (0, 1.5)]
    polygon_1 = gdstk.Polygon(points, datatype=1)
    polygon_2 = gdstk.Polygon(points).fillet(0.3, tolerance=1e-3)
    return gdstk.Cell("fillet").add(polygon_2, polygon_1)


def fracture_image():
    polygon = gdstk.racetrack((0, 0), 30, 60, 40, tolerance=1e-3)
    poly_list = polygon.fracture()
    return gdstk.Cell("fracture").add(*poly_list)


if __name__ == "__main__":
    path = pathlib.Path(__file__).parent.absolute() / "polygon"
    path.mkdir(parents=True, exist_ok=True)

    draw(init_image(), path)
    draw(bounding_box_image(), path)
    draw(fillet_image(), path)
    draw(fracture_image(), path)



# Copyright 2020 Lucas Heitzmann Gabrielli.
# This file is part of gdstk, distributed under the terms of the
# Boost Software License - Version 1.0.  See the accompanying
# LICENSE file or <http://www.boost.org/LICENSE_1_0.txt>

import subprocess
import gdstk

project = "gdstk"
version = gdstk.__version__
release = gdstk.__version__
copyright = "2020, Lucas H. Gabrielli"

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.autosummary",
    "sphinx_inline_tabs",
]

autosummary_generate = True
autosummary_imported_members = True

napoleon_google_docstring = True
napoleon_numpy_docstring = False
napoleon_include_init_with_doc = False
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = False
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = True
napoleon_use_admonition_for_references = False
napoleon_use_ivar = True
napoleon_use_param = True
napoleon_use_rtype = True

exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

html_static_path = ["_static"]

templates_path = ["_templates"]

pygments_style = "trac"

html_copy_source = False

html_show_sphinx = False

html_theme = "sphinx_rtd_theme"

html_theme_options = {
    #'canonical_url': '',
    #'analytics_id': '',
    #'logo_only': False,
    "display_version": True,
    #'prev_next_buttons_location': 'bottom',
    #'style_external_links': False,
    #'vcs_pageview_mode': '',
    "collapse_navigation": True,
    "sticky_navigation": True,
    "navigation_depth": -1,
    #'includehidden': True,
    #'titles_only': False
}

# Build images

for img_src in (
    "cell_images.py",
    "curve_images.py",
    "flexpath_images.py",
    "function_images.py",
    "label_images.py",
    "polygon_images.py",
    "reference_images.py",
    "robustpath_images.py",
    "tutorial_images.py",
    "apply_repetition.py",
    "fonts.py",
    "merging.py",
    "pads.py",
    "path_markers.py",
    "pcell.py",
    "photonics.py",
    "pos_filtering.py",
    "repetitions.py",
    "transforms.py",
    "layout.py",
    "filtering.py",
):
    subprocess.run(["python", img_src], check=True)



#!/usr/bin/env python

# Copyright 2020 Lucas Heitzmann Gabrielli.
# This file is part of gdstk, distributed under the terms of the
# Boost Software License - Version 1.0.  See the accompanying
# LICENSE file or <http://www.boost.org/LICENSE_1_0.txt>

import pathlib
import numpy
import gdstk
from tutorial_images import draw


def cross_image():
    cross1 = gdstk.cross((0, 0), 10, 1)
    cross2 = gdstk.cross((0.5, 0.5), 5, 0.75, layer=1)
    return gdstk.Cell("cross").add(cross1, cross2)


def regular_polygon_image():
    poly3 = gdstk.regular_polygon((0, 0), 9, 3)
    poly4 = gdstk.regular_polygon((10, 0), 7, 4, layer=1)
    poly5 = gdstk.regular_polygon((0, 10), 5, 5, layer=2)
    poly6 = gdstk.regular_polygon((10, 10), 4, 6, layer=3)
    return gdstk.Cell("regular_polygon").add(poly3, poly4, poly5, poly6)


def ellipse_image():
    circle = gdstk.ellipse((0, 0), 40)
    ellipse = gdstk.ellipse((100, 0), (40, 30), layer=1)
    ring = gdstk.ellipse((0, 100), 40, inner_radius=(30, 20), layer=2)
    circle_slice = gdstk.ellipse(
        (100, 100),
        40,
        initial_angle=-numpy.pi / 4,
        final_angle=5 * numpy.pi / 4,
        layer=3,
    )
    ring_slice = gdstk.ellipse(
        (50, 200), (70, 30), (60, 20), -3 * numpy.pi / 4, numpy.pi / 2, layer=4
    )
    return gdstk.Cell("ellipse").add(circle, ellipse, ring, circle_slice, ring_slice)


def racetrack_image():
    racetrack1 = gdstk.racetrack((0, 0), 8, 5)
    racetrack2 = gdstk.racetrack((18, 0), 8, 5, 3, True)
    return gdstk.Cell("racetrack").add(racetrack1, racetrack2)


def text_image():
    text = gdstk.text(f"Created with\nGDSTK {gdstk.__version__}", 1, (0, 0))
    rect = gdstk.rectangle((0, -5 / 4), (5 * 9 / 16, -1 / 4), datatype=1)
    return gdstk.Cell("text").add(*text, rect)


def contour_image():
    y, x = numpy.mgrid[0.5:1.5:128j, -1:1:255j]
    data = numpy.sin(4 * numpy.pi * x * y) - 0.8 * numpy.sin(2 * numpy.pi * y)
    contours = gdstk.contour(data, 0, 1 / 127, 0.01 / 127)
    rect = gdstk.rectangle((0, 0), (2, 1), datatype=1)
    return gdstk.Cell("contour").add(*contours, rect)


def offset_image():
    text = gdstk.text("#A", 10, (0, 0), datatype=1)
    circle = gdstk.ellipse(
        (5, 11), 5, initial_angle=0, final_angle=numpy.pi, datatype=1
    )
    path = gdstk.FlexPath([(0, -1), (5, -10), (10, -1)], 1, datatype=1)
    dilated = gdstk.offset(text + [circle, path], 0.4)
    eroded = gdstk.offset(text + [circle, path], -0.4, use_union=True, layer=1)
    return gdstk.Cell("offset").add(*text, circle, path, *dilated, *eroded)


def boolean_image():
    circle = gdstk.ellipse((0, 0), 50)
    path = gdstk.FlexPath((-50, 30), [5, 5], 10)
    path.interpolation(
        [(20, 15), (0, 0), (-20, -15), (50, -30)], angles=[0.3, None, None, None, 0.3]
    )
    text = gdstk.text("GDSTK", 40, (-2.5 * 40 * 9 / 16, -40 / 2))
    result = gdstk.boolean(circle, text + [path], "not")
    return gdstk.Cell("boolean").add(*result)


def slice_image():
    triangle = gdstk.regular_polygon((-10, 0), 8, 3)
    ring = gdstk.ellipse((10, 0), 8, 5, layer=1)
    result = gdstk.slice([triangle, ring], (-10, -5, 0, 6, 14), "x")
    assert len(result) == 6
    assert all(len(polys) == s for polys, s in zip(result, [1, 1, 0, 1, 2, 1]))
    return gdstk.Cell("slice").add(*[p for polys in result for p in polys])


def read_rawcells_example():
    cell1 = gdstk.Cell("CELL_1")
    cell1.add(gdstk.rectangle((0, 0), (2, 1)))
    cell2 = gdstk.Cell("CELL_2")
    cell2.add(gdstk.Reference(cell1, (-1, 0)))
    library = gdstk.Library()
    library.add(cell1, cell2)
    library.write_gds("test.gds")
    raw_cells = gdstk.read_rawcells("test.gds")
    assert tuple(sorted(raw_cells.keys())) == ("CELL_1", "CELL_2")
    assert len(raw_cells["CELL_1"].dependencies(True)) == 0
    assert len(raw_cells["CELL_2"].dependencies(True)) == 1
    deps = raw_cells["CELL_2"].dependencies(True)
    assert deps[0] is raw_cells["CELL_1"]


if __name__ == "__main__":
    path = pathlib.Path(__file__).parent.absolute() / "function"
    path.mkdir(parents=True, exist_ok=True)

    draw(cross_image(), path)
    draw(regular_polygon_image(), path)
    draw(ellipse_image(), path)
    draw(racetrack_image(), path)
    draw(text_image(), path)
    draw(contour_image(), path)
    draw(offset_image(), path)
    draw(boolean_image(), path)
    draw(slice_image(), path)
    read_rawcells_example()



# Copyright 2020 Lucas Heitzmann Gabrielli.
# This file is part of gdstk, distributed under the terms of the
# Boost Software License - Version 1.0.  See the accompanying
# LICENSE file or <http://www.boost.org/LICENSE_1_0.txt>

import pathlib
from tutorial_images import draw
import numpy
import gdstk


if __name__ == "__main__":
    # X-explicit repetition
    vline = gdstk.FlexPath([(3, 2), (3, 3.5)], 0.1, simple_path=True)
    vline.repetition = gdstk.Repetition(x_offsets=[0.2, 0.6, 1.4, 3.0])

    # Y-explicit repetition
    hline = gdstk.RobustPath((3, 2), 0.05, simple_path=True)
    hline.segment((6, 2))
    hline.repetition = gdstk.Repetition(y_offsets=[0.1, 0.3, 0.7, 1.5])

    # Create all copies
    vlines = vline.apply_repetition()
    hlines = hline.apply_repetition()

    # Include original elements for boolean operation
    vlines.append(vline)
    hlines.append(hline)

    result = gdstk.boolean(vlines, hlines, "or")

    main = gdstk.Cell("Main")
    main.add(*result)
    main.name = "apply_repetition"
    path = pathlib.Path(__file__).parent.absolute() / "how-tos"
    path.mkdir(parents=True, exist_ok=True)
    draw(main, path)



# Copyright 2020 Lucas Heitzmann Gabrielli.
# This file is part of gdstk, distributed under the terms of the
# Boost Software License - Version 1.0.  See the accompanying
# LICENSE file or <http://www.boost.org/LICENSE_1_0.txt>

from tutorial_images import draw
import pathlib
import numpy
import gdstk
import pcell


if __name__ == "__main__":
    path = pathlib.Path(__file__).parent.absolute()

    # Check library units. In this case it is using the default units.
    units = gdstk.gds_units(path / "photonics.gds")
    print(f"Using unit = {units[0]}, precision = {units[1]}")

    # Load the library as a dictionary of RawCell
    pdk = gdstk.read_rawcells(path / "photonics.gds")

    # Cell holding a single device (MZI)
    dev_cell = gdstk.Cell("Device")
    dev_cell.add(gdstk.Reference(pdk["MZI"], (-40, 0)))

    # Create a grating coupler using the function imported from the
    # pcell module (pcell.py) created earlier.
    grating = pcell.grating(0.62, layer=2)
    # Add 4 grating couplers to the device cell: one for each port.
    dev_cell.add(
        gdstk.Reference(
            grating, (-200, -150), rotation=numpy.pi / 2, columns=2, spacing=(300, 0)
        ),
        gdstk.Reference(
            grating, (200, 150), rotation=-numpy.pi / 2, columns=2, spacing=(300, 0)
        ),
    )

    # Create a waveguide connecting a grating to a MZI port.
    waveguide = gdstk.FlexPath((-220, -150), 20, bend_radius=15, layer=1)
    # Grating background
    waveguide.segment((20, 0), relative=True)
    # Linear taper
    waveguide.segment((-100, -150), 0.5)
    # Connection to MZI
    waveguide.segment([(-70, -150), (-70, -1), (-40, -1)])
    # Since the device is symmetrical, we can create a cell with the
    # waveguide geometry and reuse it for all 4 ports.
    wg_cell = gdstk.Cell("Waveguide")
    wg_cell.add(waveguide)

    dev_cell.add(gdstk.Reference(wg_cell))
    dev_cell.add(gdstk.Reference(wg_cell, x_reflection=True))
    dev_cell.add(gdstk.Reference(wg_cell, rotation=numpy.pi))
    dev_cell.add(gdstk.Reference(wg_cell, rotation=numpy.pi, x_reflection=True))

    # Main cell with 2 devices and lithography alignment marks
    main = gdstk.Cell("Main")
    main.add(
        gdstk.Reference(dev_cell, (250, 250)),
        gdstk.Reference(dev_cell, (250, 750)),
        gdstk.Reference(pdk["Alignment Mark"], columns=2, rows=3, spacing=(500, 500)),
    )

    lib = gdstk.Library()
    lib.add(main, *main.dependencies(True))
    lib.write_gds(path / "layout.gds")

    pdk_lib = gdstk.read_gds(path / "photonics.gds")
    pdk = {c.name: c for c in pdk_lib.cells}
    for cell in [dev_cell, main]:
        for x in cell.references:
            if isinstance(x.cell, gdstk.RawCell):
                ref = gdstk.Reference(pdk[x.cell.name], x.origin)
                ref.repetition = x.repetition
                cell.remove(x)
                cell.add(ref)
    main.name = "layout"
    draw(main, path / "how-tos")



#!/usr/bin/env python

# Copyright 2020 Lucas Heitzmann Gabrielli.
# This file is part of gdstk, distributed under the terms of the
# Boost Software License - Version 1.0.  See the accompanying
# LICENSE file or <http://www.boost.org/LICENSE_1_0.txt>

import pathlib
import numpy
import gdstk
from tutorial_images import draw


def init_image():
    path = gdstk.FlexPath(
        [(0, 5), (0, 0), (5, 0), (15, 10), (15, -5)],
        [0.8, 0.8, 0.8, 0.8],
        1.0,
        joins=["natural", "bevel", "miter", "round"],
        ends=["flush", "extended", (0.4, 0.8), "round"],
        layer=[0, 1, 2, 3],
    )
    return gdstk.Cell("init").add(path)


def init0_image():
    points = [(0, 8), (0, 0), (8, 0), (18, 13), (18, -8)]
    path_1 = gdstk.FlexPath(points, 1, datatype=1)
    path_2 = gdstk.FlexPath(points, 1, bend_radius=3)
    return gdstk.Cell("init0").add(path_2, path_1)


def init1_image():
    def custom_broken_join(p0, v0, p1, v1, center, width):
        p0 = numpy.array(p0)
        v0 = numpy.array(v0)
        p1 = numpy.array(p1)
        v1 = numpy.array(v1)
        center = numpy.array(center)
        # Calculate intersection point p between lines defined by
        # p0 + u0 * v0 (for all u0) and p1 + u1 * v1 (for all u1)
        den = v1[1] * v0[0] - v1[0] * v0[1]
        lim = 1e-12 * (v0[0] ** 2 + v0[1] ** 2) * (v1[0] ** 2 + v1[1] ** 2)
        if den**2 < lim:
            # Lines are parallel: use mid-point
            u0 = u1 = 0
            p = 0.5 * (p0 + p1)
        else:
            dx = p1[0] - p0[0]
            dy = p1[1] - p0[1]
            u0 = (v1[1] * dx - v1[0] * dy) / den
            u1 = (v0[1] * dx - v0[0] * dy) / den
            p = 0.5 * (p0 + v0 * u0 + p1 + v1 * u1)
        if u0 <= 0 and u1 >= 0:
            # Inner corner
            return [p]
        # Outer corner
        return [p0, center, p1]

    def custom_pointy_end(p0, v0, p1, v1):
        p0 = numpy.array(p0)
        v0 = numpy.array(v0)
        p1 = numpy.array(p1)
        v1 = numpy.array(v1)
        r = 0.5 * numpy.sqrt(numpy.sum((p0 - p1) ** 2))
        v0 /= numpy.sqrt(numpy.sum(v0**2))
        v1 /= numpy.sqrt(numpy.sum(v1**2))
        return [p0, 0.5 * (p0 + p1) + 0.5 * (v0 - v1) * r, p1]

    path = gdstk.FlexPath(
        [(0, 5), (0, 0), (5, 0), (15, 10), (15, -5)],
        3,
        joins=custom_broken_join,
        ends=custom_pointy_end,
    )
    return gdstk.Cell("init1").add(path)


def horizontal_image():
    path = gdstk.FlexPath((0, 0), 0.2)
    path.horizontal(2, width=0.4, relative=True)
    path.horizontal(2, offset=[0.4], relative=True)
    path.horizontal(2, relative=True)
    assert (
        numpy.max(
            numpy.abs(
                path.spine()
                - numpy.array([[0.0, 0.0], [2.0, 0.0], [4.0, 0.0], [6.0, 0.0]])
            )
        )
        == 0
    )
    return gdstk.Cell("horizontal").add(path)


def segment_image():
    points = [(1, 0), (1, 1), (-1, 1), (-1, -1), (1, -1)]
    path_1 = gdstk.FlexPath((0, 0), 0.2)
    path_1.segment(points, 0.6)
    path_2 = gdstk.FlexPath((3, 0), [0.1, 0.1], 0.2)
    path_2.segment(points, offset=0.6, relative=True)
    return gdstk.Cell("segment").add(path_1, path_2)


def cubic_image():
    path = gdstk.FlexPath((0, 0), 0.2, tolerance=1e-3)
    path.cubic([(0, 1), (1, 1), (1, 0)])
    path.cubic([(1, -1), (2, -1), (2.5, -0.5), (3, 0), (3, 1), (2, 1)], width=0.5)
    return gdstk.Cell("cubic").add(path)


def cubic_smooth_image():
    path = gdstk.FlexPath((0, 0), 0.2, tolerance=1e-3)
    path.cubic([(0, 1), (1, 1), (1, 0)])
    path.cubic_smooth([(2, -1), (2.5, -0.5), (3, 1), (2, 1)], width=0.5)
    return gdstk.Cell("cubic_smooth").add(path)


def bezier_image():
    path = gdstk.FlexPath((0, 0), 0.2, tolerance=1e-3)
    path.bezier([(4, 1), (4, 3), (0, 5), (-4, 3), (-4, -2), (0, -4)])
    return gdstk.Cell("bezier").add(path)


def interpolation_image():
    half_pi = numpy.pi / 2
    points = [(4, 1), (4, 3), (0, 5), (-4, 3), (-4, -2), (0, -4)]
    angles = [half_pi, None, None, None, -half_pi, -half_pi, None]
    path_1 = gdstk.FlexPath((0, 0), 0.2, tolerance=1e-3)
    path_1.interpolation(points, cycle=True)
    path_2 = gdstk.FlexPath((6, -8), 0.2, tolerance=1e-3)
    path_2.interpolation(points, angles, cycle=True, relative=True)
    return gdstk.Cell("interpolation").add(path_1, path_2)


def arc_image():
    path = gdstk.FlexPath((0, 0), [0.2, 0.3], 0.4, tolerance=1e-3)
    path.vertical(5)
    path.arc(2.5, numpy.pi, 0)
    path.arc(5, -numpy.pi, -numpy.pi / 2)
    return gdstk.Cell("arc").add(path)


def parametric_image():
    def spiral(u):
        rad = 2 * u**0.5
        ang = 3 * numpy.pi * u
        return (rad * numpy.cos(ang), rad * numpy.sin(ang))

    path = gdstk.FlexPath((0, 0), 0.2, tolerance=1e-3)
    path.parametric(spiral)
    return gdstk.Cell("parametric").add(path)


def commands_image():
    path = gdstk.FlexPath((0, 0), [0.2, 0.4, 0.2], 0.5, tolerance=1e-3)
    path.commands(
        "l",
        3,
        4,
        "A",
        2,
        numpy.arctan2(3, -4),
        numpy.pi / 2,
        "h",
        0.5,
        "a",
        3,
        -numpy.pi,
    )
    return gdstk.Cell("commands").add(path)


if __name__ == "__main__":
    path = pathlib.Path(__file__).parent.absolute() / "flexpath"
    path.mkdir(parents=True, exist_ok=True)

    draw(init_image(), path)
    draw(init0_image(), path)
    draw(init1_image(), path)
    draw(horizontal_image(), path)
    draw(segment_image(), path)
    draw(cubic_image(), path)
    draw(cubic_smooth_image(), path)
    draw(bezier_image(), path)
    draw(interpolation_image(), path)
    draw(arc_image(), path)
    draw(parametric_image(), path)
    draw(commands_image(), path)



#!/usr/bin/env python

# Copyright 2020 Lucas Heitzmann Gabrielli.
# This file is part of gdstk, distributed under the terms of the
# Boost Software License - Version 1.0.  See the accompanying
# LICENSE file or <http://www.boost.org/LICENSE_1_0.txt>

import pathlib
import numpy
import gdstk
from tutorial_images import draw


def bounding_box_image():
    polygons = gdstk.text("F", 10, (0, 0))
    f_cell = gdstk.Cell("F_CELL")
    f_cell.add(*polygons)
    array_ref = gdstk.Reference(
        f_cell, rotation=numpy.pi / 4, columns=3, rows=2, spacing=(8, 10)
    )
    path = gdstk.FlexPath([(-5, 0), (0, -5), (5, 0)], 1, simple_path=True)
    main_cell = gdstk.Cell("MAIN")
    main_cell.add(array_ref, path)
    bbox = main_cell.bounding_box()
    assert bbox == (
        (-12.816310409006173, -5.707106781186548),
        (11.313708498984761, 27.66555281392367),
    )
    polygon_bb = gdstk.rectangle(*bbox, datatype=1)
    main_cell.name = "bounding_box"
    return main_cell.add(polygon_bb)


def convex_hull_image():
    polygons = gdstk.text("F", 10, (0, 0))
    f_cell = gdstk.Cell("F_CELL")
    f_cell.add(*polygons)
    array_ref = gdstk.Reference(
        f_cell, rotation=numpy.pi / 4, columns=3, rows=2, spacing=(8, 10)
    )
    path = gdstk.FlexPath([(-5, 0), (0, -5), (5, 0)], 1, simple_path=True)
    main_cell = gdstk.Cell("MAIN")
    main_cell.add(array_ref, path)
    hull = main_cell.convex_hull()
    error = hull - numpy.array(
        [
            [1.14904852, 27.66555281],
            [-12.81631041, 13.70019389],
            [-5.35355339, -0.35355339],
            [0.0, -5.70710678],
            [5.35355339, -0.35355339],
            [11.3137085, 13.96535893],
            [9.98788328, 17.94283457],
            [8.66205807, 20.15254326],
        ]
    )
    assert numpy.abs(error).max() < 1e-8
    polygon_hull = gdstk.Polygon(hull, datatype=1)
    main_cell.name = "convex_hull"
    return main_cell.add(polygon_hull)


def flatten_image():
    poly1 = gdstk.Polygon([(0, 0), (1, 0), (0.5, 1)])
    cell1 = gdstk.Cell("CELL_1")
    cell1.add(poly1)
    poly2 = gdstk.Polygon([(1, 0), (1.5, 1), (0.5, 1)], layer=1)
    ref = gdstk.Reference(cell1, columns=2, rows=2, spacing=(1, 1))
    cell2 = gdstk.Cell("CELL_2")
    cell2.add(poly2, ref)
    assert len(cell2.polygons) == 1
    assert len(cell2.references) == 1
    assert len(cell2.dependencies(True)) == 1
    cell2.flatten()
    assert len(cell2.polygons) == 5
    assert len(cell2.references) == 0
    assert len(cell2.dependencies(True)) == 0
    cell2.name = "flatten"
    return cell2


def write_svg_image():
    # (layer, datatype) = (0, 1)
    poly1 = gdstk.ellipse((0, 0), (13, 10), datatype=1)
    # (layer, datatype) = (1, 0)
    poly2 = gdstk.ellipse((0, 0), (10, 7), layer=1)
    (layer, texttype) = (3, 2)
    label = gdstk.Label("Example", (0, 0), layer=3, texttype=2)
    cell = gdstk.Cell("SVG")
    cell.add(poly1, poly2, label)
    # cell.write_svg(
    #     "example.svg",
    #     background="none",
    #     shape_style={(0, 1): {"fill": "none", "stroke": "black", "stroke-dasharray": "8,8"}},
    #     label_tyle={(3, 2): {"fill": "none", "stroke": "red", "font-size": "32px"}},
    #     pad="5%",
    #     sort_function=lambda p1, p2: p1.layer < p2.layer,
    # )
    cell.name = "write_svg"
    return cell


def remove_image():
    polygons = gdstk.text("Filter dots\nin i and j!", 8, (0, 0))
    cell = gdstk.Cell("FILTERED")
    cell.add(*polygons)
    dots = [poly for poly in cell.polygons if poly.area() < 2]
    cell.remove(*dots)
    cell.name = "remove"
    return cell


if __name__ == "__main__":
    path = pathlib.Path(__file__).parent.absolute() / "cell"
    path.mkdir(parents=True, exist_ok=True)

    draw(bounding_box_image(), path)
    draw(convex_hull_image(), path)
    draw(flatten_image(), path)
    draw(write_svg_image(), path)
    draw(remove_image(), path)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2020 Lucas Heitzmann Gabrielli.
# This file is part of gdstk, distributed under the terms of the
# Boost Software License - Version 1.0.  See the accompanying
# LICENSE file or <http://www.boost.org/LICENSE_1_0.txt>

import numpy
import pathlib
import platform
import setuptools
from setuptools.command.build_ext import build_ext


class CMakeBuilder(build_ext):
    def run(self):
        darwin = platform.system() == "Darwin"
        root_dir = pathlib.Path().absolute()
        build_dir = pathlib.Path(self.build_temp).absolute() / "cmake_build"
        install_dir = build_dir / "install"

        config = "Debug" if self.debug else "Release"

        build_dir.mkdir(parents=True, exist_ok=True)
        self.spawn(
            [
                "cmake",
                "-S",
                str(root_dir),
                "-B",
                str(build_dir),
                "-DCMAKE_INSTALL_PREFIX=" + str(install_dir),
                "-DCMAKE_BUILD_TYPE=" + config,
            ]
        )
        if not self.dry_run:
            self.spawn(
                [
                    "cmake",
                    "--build",
                    str(build_dir),
                    "--config",
                    config,
                    "--target",
                    "install",
                ]
            )

        pkgconfig = list(install_dir.glob("**/gdstk.pc"))
        if len(pkgconfig) == 0:
            raise RuntimeError(f"File gdstk.pc not found in cmake install tree: {install_dir}")

        with open(pkgconfig[0]) as pkg:
            for line in pkg:
                if line.startswith("Cflags:"):
                    for arg in line.split()[1:]:
                        if arg.startswith("-I"):
                            self.extensions[0].include_dirs.append(arg[2:])
                        else:
                            self.extensions[0].extra_compile_args.append(arg)
                elif line.startswith("Libs:"):
                    for arg in line.split()[1:]:
                        if darwin and (arg == "-lpython" or arg.endswith("Python.framework")):
                            # Do not link to python in macOS, we use the -undefined dynamic_lookup
                            # linker flag instead.
                            # See https://blog.tim-smith.us/2015/09/python-extension-modules-os-x/
                            continue
                        if arg.endswith(".framework"):
                            # macOS-specific
                            self.extensions[0].extra_link_args.extend(
                                ["-framework", arg[arg.rfind("/") + 1 : -10]]
                            )
                        elif arg.startswith("-L"):
                            self.extensions[0].library_dirs.append(arg[2:])
                        elif arg.startswith("-l"):
                            self.extensions[0].libraries.append(arg[2:])
                        else:
                            self.extensions[0].extra_link_args.append(arg)

        super().run()


extra_compile_args = []
extra_link_args = []
if platform.system() == "Darwin":
    extra_compile_args.extend(["-std=c++11", "-mmacosx-version-min=10.9"])
    extra_link_args.extend(["-undefined", "dynamic_lookup", "-flat_namespace"])

setuptools.setup(
    ext_modules=[
        setuptools.Extension(
            "gdstk.gdstk",
            ["python/gdstk_module.cpp"],
            include_dirs=[numpy.get_include()],
            extra_compile_args=extra_compile_args,
            extra_link_args=extra_link_args,
        ),
    ],
    packages=["gdstk"],
    package_data={"gdstk": ["*.pyi", "py.typed"]},
    include_package_data=True,
    cmdclass={"build_ext": CMakeBuilder},
    zip_safe=False,
)



#!/usr/bin/env python

# Copyright 2020 Lucas Heitzmann Gabrielli.
# This file is part of gdstk, distributed under the terms of the
# Boost Software License - Version 1.0.  See the accompanying
# LICENSE file or <http://www.boost.org/LICENSE_1_0.txt>

from copy import deepcopy
import numpy
import gdstk

from conftest import assert_same_shape


def test_init():
    path = gdstk.FlexPath(0j, 2)
    assert path.layers == (0,)
    assert path.datatypes == (0,)
    assert path.num_paths == 1
    assert path.size == 1
    assert path.tolerance == 1e-2
    assert not path.simple_path
    assert path.scale_width
    assert path.joins == ("natural",)
    assert path.ends == ("flush",)
    assert path.bend_radius == (0,)
    assert path.bend_function == (None,)
    path.set_layers(1)
    path.set_datatypes(2)
    path.tolerance = 1e-3
    path.simple_path = True
    path.set_joins("round")
    path.set_ends((1, 2))
    path.set_bend_radius(0.1)
    path.set_bend_function(None)
    assert path.layers == (1,)
    assert path.datatypes == (2,)
    assert path.tolerance == 1e-3
    assert path.simple_path
    assert path.joins == ("round",)
    assert path.ends == ((1, 2),)
    assert path.bend_radius == (0.1,)
    assert path.bend_function == (None,)

    path = gdstk.FlexPath(0j, [2, 2], layer=3, datatype=[4, 5])
    assert path.layers == (3, 3)
    assert path.datatypes == (4, 5)
    assert path.num_paths == 2
    assert path.size == 1
    path.set_layers(5, 4)
    path.set_datatypes(3, 2)
    assert path.layers == (5, 4)
    assert path.datatypes == (3, 2)

    path = gdstk.FlexPath((0j, (1, 1)), 2, [-1, 1])
    assert path.layers == (0, 0)
    assert path.datatypes == (0, 0)
    assert path.num_paths == 2

    path = gdstk.FlexPath((1j, -1j), [2, 2], [-1, 1], layer=3, datatype=[4, 5])
    assert path.layers == (3, 3)
    assert path.datatypes == (4, 5)
    assert path.num_paths == 2
    assert path.bend_function == (None, None)
    dummy = lambda *args: (0j,)  # noqa E731
    path.set_bend_function(dummy, None)
    assert path.bend_function == (dummy, None)


def test_transforms(proof_cells):
    for scale_width in [True, False]:
        path0 = gdstk.FlexPath(
            (
                0j,
                1j,
                0.5 + 1j,
            ),
            [0.1, 0.2],
            0.3,
            tolerance=1e-4,
            scale_width=scale_width,
        )
        path0.turn(0.4, -numpy.pi, [0.2, 0.1]).segment((-0.2, 0), relative=True)
        paths = [
            path0,
            path0.copy().mirror((1.5, 0)),
            path0.copy().mirror((1.5, 0), (1.5, 1)),
            path0.copy().scale(2, (3, 0)),
            path0.copy().scale(-2, (-1, 0)),
            path0.copy().rotate(numpy.pi / 2, (2, 1)).translate(0.2, -0.3),
        ]
        assert_same_shape(proof_cells[f"FlexPath:scale_width_{scale_width}"].polygons, paths)


def test_points():
    path = gdstk.FlexPath((0j, 10j), 2).horizontal(10, 2, 1).vertical(0, 1, [1])
    numpy.testing.assert_array_equal(path.spine(), [[0, 0], [0, 10], [10, 10], [10, 0]])
    numpy.testing.assert_array_equal(path.widths(), [[2], [2], [2], [1]])
    numpy.testing.assert_array_equal(path.offsets(), [[0], [0], [0], [1]])
    path_spines = path.path_spines()
    assert len(path_spines) == 1
    numpy.testing.assert_array_equal(path_spines[0], [[0, 0], [0, 10], [10, 10], [11, 0]])

    path = gdstk.FlexPath((0j, 10j), [2, 2], 2).horizontal(10, 2, 1).vertical(0, 1, [-2, 1])
    numpy.testing.assert_array_equal(path.spine(), [[0, 0], [0, 10], [10, 10], [10, 0]])
    numpy.testing.assert_array_equal(path.widths(), [[2, 2], [2, 2], [2, 2], [1, 1]])
    numpy.testing.assert_array_equal(path.offsets(), [[-1, 1], [-1, 1], [-0.5, 0.5], [-2, 1]])
    path_spines = path.path_spines()
    assert len(path_spines) == 2


def test_deepcopy():
    path = gdstk.FlexPath(0j, 2)
    path2 = deepcopy(path)

    assert path is not path2

    assert path.layers == path2.layers
    assert path.datatypes == path2.datatypes
    assert path.num_paths == path2.num_paths
    assert path.size == path2.size
    assert path.tolerance == path2.tolerance
    assert path.simple_path == path2.simple_path
    assert path.scale_width == path2.scale_width
    assert path.joins == path2.joins
    assert path.ends == path2.ends
    assert path.bend_radius == path2.bend_radius
    assert path.bend_function == path2.bend_function

    path2.set_layers(1)
    assert path.layers == (0,)
    assert path2.layers == (1,)



#!/usr/bin/env python

# Copyright 2020 Lucas Heitzmann Gabrielli.
# This file is part of gdstk, distributed under the terms of the
# Boost Software License - Version 1.0.  See the accompanying
# LICENSE file or <http://www.boost.org/LICENSE_1_0.txt>

import pathlib
import numpy
import gdstk


def draw(cell, path):
    bb = cell.bounding_box()
    scaling = 300 / (1.1 * (bb[1][0] - bb[0][0]))
    name = path / (cell.name + ".svg")
    cell.write_svg(
        name,
        scaling=scaling,
        background="none",
        shape_style={
            (0, 1): {"fill": "none", "stroke": "black", "stroke-dasharray": "8,8"}
        },
        label_style={(3, 2): {"stroke": "red", "fill": "none", "font-size": "32px"}},
        pad="5%",
    )
    print(f"Saving {name} (scaling {scaling})")


if __name__ == "__main__":
    path = pathlib.Path(__file__).parent.absolute() / "tutorial"
    path.mkdir(parents=True, exist_ok=True)

    # Polygons
    # Create a polygon from a list of vertices
    points = [(0, 0), (2, 2), (2, 6), (-6, 6), (-6, -6), (-4, -4), (-4, 4), (0, 4)]
    poly = gdstk.Polygon(points)
    draw(gdstk.Cell("polygons").add(poly), path)

    # Holes
    # Manually connect the hole to the outer boundary
    cutout = gdstk.Polygon(
        [(0, 0), (5, 0), (5, 5), (0, 5), (0, 0), (2, 2), (2, 3), (3, 3), (3, 2), (2, 2)]
    )
    draw(gdstk.Cell("holes").add(cutout), path)

    # Circles
    # Circle centered at (0, 0), with radius 2 and tolerance 0.1
    circle = gdstk.ellipse((0, 0), 2, tolerance=0.01)

    # To create an ellipse, simply pass a list with 2 radii.
    # Because the tolerance is small (resulting a large number of
    # vertices), the ellipse is fractured in 2 polygons.
    ellipse = gdstk.ellipse((4, 0), [1, 2], tolerance=1e-4)

    # Circular arc example
    arc = gdstk.ellipse(
        (2, 4),
        2,
        inner_radius=1,
        initial_angle=-0.2 * numpy.pi,
        final_angle=1.2 * numpy.pi,
        tolerance=0.01,
    )
    draw(gdstk.Cell("circles").add(circle, ellipse, arc), path)

    # Curves
    # Construct a curve made of a sequence of line segments
    c1 = gdstk.Curve((0, 0)).segment([(1, 0), (2, 1), (2, 2), (0, 2)])
    p1 = gdstk.Polygon(c1.points())

    # Construct another curve using relative coordinates
    c2 = gdstk.Curve((3, 1)).segment([(1, 0), (2, 1), (2, 2), (0, 2)], relative=True)
    p2 = gdstk.Polygon(c2.points())
    draw(gdstk.Cell("curves").add(p1, p2), path)

    # Curves 1
    # Use complex numbers to facilitate writing polar coordinates
    c3 = gdstk.Curve(2j).segment(4 * numpy.exp(1j * numpy.pi / 6), relative=True)
    # Elliptical arcs have syntax similar to gdstk.ellipse
    c3.arc((4, 2), 0.5 * numpy.pi, -0.5 * numpy.pi)
    p3 = gdstk.Polygon(c3.points())
    draw(gdstk.Cell("curves_1").add(p3), path)

    # Curves 2
    # Cubic Bezier curves can be easily created
    c4 = gdstk.Curve((0, 0), tolerance=1e-3)
    c4.cubic([(0, 1), (1, 1), (1, 0)])
    # Smooth continuation:
    c4.cubic_smooth([(1, -1), (1, 0)], relative=True)

    # Similarly for quadratic Bezier curves
    c4.quadratic([(0.5, 1), (1, 0)], relative=True)
    c4.quadratic_smooth((1, 0), relative=True)

    # Smooth interpolating curve
    c4.interpolation([(4, -1), (3, -2), (2, -1.5), (1, -2), (0, -1), (0, 0)])
    p4 = gdstk.Polygon(c4.points())
    # draw
    ref_poly = gdstk.Polygon(
        [
            0j,
            1j,
            1 + 1j,
            1 + 0j,
            1 - 1j,
            2 - 1j,
            2 + 0j,
            2.5 + 1j,
            3 + 0j,
            3.5 - 1j,
            4 + 0j,
            4 - 1j,
            3 - 2j,
            2 - 1.5j,
            1 - 2j,
            -1j,
        ],
        datatype=1,
    )
    draw(gdstk.Cell("curves_2").add(p4, ref_poly), path)

    # Transformations
    poly = gdstk.rectangle((-2, -2), (2, 2))
    poly.rotate(numpy.pi / 4)
    poly.scale(1, 0.5)
    draw(gdstk.Cell("transformations").add(poly), path)

    # Layer and Datatype
    # Layer/datatype definitions for each step in the fabrication
    ld = {
        "full etch": {"layer": 1, "datatype": 3},
        "partial etch": {"layer": 2, "datatype": 3},
        "lift-off": {"layer": 0, "datatype": 7},
    }

    p1 = gdstk.rectangle((-3, -3), (3, 3), **ld["full etch"])
    p2 = gdstk.rectangle((-5, -3), (-3, 3), **ld["partial etch"])
    p3 = gdstk.rectangle((5, -3), (3, 3), **ld["partial etch"])
    p4 = gdstk.regular_polygon((0, 0), 2, 6, **ld["lift-off"])
    draw(gdstk.Cell("layer_and_datatype").add(p1, p2, p3, p4), path)

    # References
    # Create a cell with a component that is used repeatedly
    contact = gdstk.Cell("CONTACT")
    contact.add(p1, p2, p3, p4)

    # Create a cell with the complete device
    device = gdstk.Cell("DEVICE")
    device.add(cutout)
    # Add 2 references to the component changing size and orientation
    ref1 = gdstk.Reference(contact, (3.5, 1), magnification=0.25)
    ref2 = gdstk.Reference(contact, (1, 3.5), magnification=0.25, rotation=numpy.pi / 2)
    device.add(ref1, ref2)

    # The final layout has several repetitions of the complete device
    main = gdstk.Cell("MAIN")
    main.add(gdstk.Reference(device, (0, 0), columns=3, rows=2, spacing=(6, 7)))
    # draw
    main.name = "references"
    draw(main, path)
    main.name = "MAIN"

    # Flexible Paths
    # Path defined by a sequence of points and stored as a GDSII path
    fp1 = gdstk.FlexPath(
        [(0, 0), (3, 0), (3, 2), (5, 3), (3, 4), (0, 4)], 1, simple_path=True
    )

    # Other construction methods can still be used
    fp1.interpolation([(0, 2), (2, 2), (4, 3), (5, 1)], relative=True)

    # Multiple parallel paths separated by 0.5 with different widths,
    # end caps, and joins.  Because of the join specification, they
    # cannot be stared as GDSII paths, only as polygons.
    fp2 = gdstk.FlexPath(
        [(12, 0), (8, 0), (8, 3), (10, 2)],
        [0.3, 0.2, 0.4],
        0.5,
        ends=["extended", "flush", "round"],
        joins=["bevel", "miter", "round"],
    )
    fp2.arc(2, -0.5 * numpy.pi, 0.5 * numpy.pi)
    fp2.arc(1, 0.5 * numpy.pi, 1.5 * numpy.pi)
    draw(gdstk.Cell("flexible_paths").add(fp1, fp2), path)

    # Flexible Paths 1
    # Path created with automatic bends of radius 5
    points = [(0, 0), (0, 10), (20, 0), (18, 15), (8, 15)]
    fp3 = gdstk.FlexPath(points, 0.5, bend_radius=5, simple_path=True)

    # Same path, generated with natural joins, for comparison
    fp4 = gdstk.FlexPath(points, 0.5, layer=1, simple_path=True)
    draw(gdstk.Cell("flexible_paths_2").add(fp3, fp4), path)

    # Flexible Paths 2
    # Straight segment showing the possibility of width and offset changes
    fp5 = gdstk.FlexPath((0, 0), [0.5, 0.5], 1)
    fp5.horizontal(2)
    fp5.horizontal(4, width=0.8, offset=1.8)
    fp5.horizontal(6)
    draw(gdstk.Cell("flexible_paths_3").add(fp5), path)

    # Robust Paths
    # Create 4 parallel paths in different layers
    rp = gdstk.RobustPath(
        (0, 50),
        [2, 0.5, 1, 1],
        [0, 0, -1, 1],
        ends=["extended", "round", "flush", "flush"],
        layer=[1, 0, 2, 2],
    )
    rp.segment((0, 45))
    rp.segment(
        (0, 5),
        width=[lambda u: 2 + 16 * u * (1 - u), 0.5, 1, 1],
        offset=[
            0,
            lambda u: 8 * u * (1 - u) * numpy.cos(12 * numpy.pi * u),
            lambda u: -1 - 8 * u * (1 - u),
            lambda u: 1 + 8 * u * (1 - u),
        ],
    )
    rp.segment((0, 0))
    rp.interpolation(
        [(15, 5)],
        angles=[0, 0.5 * numpy.pi],
        width=0.5,
        offset=[-0.25, 0.25, -0.75, 0.75],
    )
    rp.parametric(
        lambda u: numpy.array((4 * numpy.sin(6 * numpy.pi * u), 45 * u)),
        offset=[
            lambda u: -0.25 * numpy.cos(24 * numpy.pi * u),
            lambda u: 0.25 * numpy.cos(24 * numpy.pi * u),
            -0.75,
            0.75,
        ],
    )
    draw(gdstk.Cell("robust_paths").add(rp), path)

    # Text
    # Label centered at (1, 3)
    label = gdstk.Label("Sample label", (5, 3), texttype=2)

    # Horizontal text with height 2.25
    htext = gdstk.text("12345", 2.25, (0.25, 6))

    # Vertical text with height 1.5
    vtext = gdstk.text("ABC", 1.5, (10.5, 4), vertical=True)

    rect = gdstk.rectangle((0, 0), (10, 6), layer=10)
    draw(gdstk.Cell("text").add(*htext, *vtext, label, rect), path)

    # Boolean Operations
    # Create some text
    text = gdstk.text("GDSTK", 4, (0, 0))
    # Create a rectangle extending the text's bounding box by 1
    rect = gdstk.rectangle((-1, -1), (5 * 4 * 9 / 16 + 1, 4 + 1))

    # Subtract the text from the rectangle
    inv = gdstk.boolean(rect, text, "not")
    draw(gdstk.Cell("boolean_operations").add(*inv), path)

    # Slice Operation
    ring1 = gdstk.ellipse((-6, 0), 6, inner_radius=4)
    ring2 = gdstk.ellipse((0, 0), 6, inner_radius=4)
    ring3 = gdstk.ellipse((6, 0), 6, inner_radius=4)

    # Slice the first ring across x=-3, the second ring across x=-3
    # and x=3, and the third ring across x=3
    slices1 = gdstk.slice(ring1, -3, "x")
    slices2 = gdstk.slice(ring2, [-3, 3], "x")
    slices3 = gdstk.slice(ring3, 3, "x")

    slices = gdstk.Cell("SLICES")

    # Keep only the left side of slices1, the center part of slices2
    # and the right side of slices3
    slices.add(*slices1[0])
    slices.add(*slices2[1])
    slices.add(*slices3[1])
    # draw
    slices.name = "slice_operation"
    draw(slices, path)
    slices.name = "SLICES"

    # Offset Operation
    rect1 = gdstk.rectangle((-4, -4), (1, 1))
    rect2 = gdstk.rectangle((-1, -1), (4, 4))

    # Erosion: because we set `use_union=True`, the inner boundaries have no effect
    outer = gdstk.offset([rect1, rect2], -0.5, use_union=True, layer=1)
    draw(gdstk.Cell("offset_operation").add(rect1, rect2, *outer), path)

    # Fillet Operation
    flexpath = gdstk.FlexPath([(-8, -4), (0, -4), (0, 4), (8, 4)], 4)
    filleted_path = flexpath.to_polygons()[0]
    filleted_path.fillet(1.5)
    draw(gdstk.Cell("fillet_operation").add(filleted_path), path)



#!/usr/bin/env python

# Copyright 2020 Lucas Heitzmann Gabrielli.
# This file is part of gdstk, distributed under the terms of the
# Boost Software License - Version 1.0.  See the accompanying
# LICENSE file or <http://www.boost.org/LICENSE_1_0.txt>

import gdstk


def test_gds_properties():
    for obj in [
        gdstk.Polygon([-1 + 0j, -2j, 3 + 0j, 4j]),
        gdstk.FlexPath((0j, 1j), 0.1),
        gdstk.RobustPath(0j, 0.1),
        gdstk.Label("Label", 0j),
        gdstk.Reference("EMPTY"),
    ]:
        assert obj.get_gds_property(12) is None
        assert obj.delete_gds_property(12) is obj
        obj.set_gds_property(13, "Property text")
        assert obj.get_gds_property(12) is None
        assert obj.get_gds_property(13) == "Property text"
        obj.delete_gds_property(13)
        assert obj.get_gds_property(13) is None
        obj.set_gds_property(13, "Second text")
        obj.set_gds_property(13, "Third text")
        obj.set_gds_property(14, "Fourth text")
        assert obj.get_gds_property(13) == "Third text"
        assert obj.properties == [
            ["S_GDS_PROPERTY", 14, b"Fourth text\x00"],
            ["S_GDS_PROPERTY", 13, b"Third text\x00"],
        ]


def test_properties():
    for obj in [
        gdstk.Polygon([-1 + 0j, -2j, 3 + 0j, 4j]),
        gdstk.FlexPath((0j, 1j), 0.1),
        gdstk.RobustPath(0j, 0.1),
        gdstk.Label("Label", 0j),
        gdstk.Reference("EMPTY"),
        gdstk.Cell("CELL"),
        gdstk.Library("Name"),
    ]:
        assert len(obj.properties) == 0
        assert obj.get_property("None") is None
        obj.set_property("FIRST", 1)
        obj.set_property("SECOND", 2.0)
        obj.set_property("THIRD", -3)
        obj.set_property("FOURTH", [1, 2.0, -3, "FO", b"UR\x00TH\x00"])
        obj.set_property("FIRST", -1)
        assert obj.get_property("FIRST") == [-1]
        obj.delete_property("THIRD")
        assert obj.properties == [
            ["FIRST", -1],
            ["FOURTH", 1, 2.0, -3, b"FO", b"UR\x00TH\x00"],
            ["SECOND", 2.0],
            ["FIRST", 1],
        ]
        obj.properties = (
            ("ONE", -1),
            ("TWO", -2.3e-4, "two"),
            ("Three", b"\xFF\xEE", 0),
        )
        assert obj.properties == [
            ["ONE", -1],
            ["TWO", -2.3e-4, b"two"],
            ["Three", b"\xFF\xEE", 0],
        ]
